"""
formatters.py

Output formatters to write HTML representations of data stored in variables
of various types.
"""

# We will use the methods of minidom.Text objects to encode text for
# displaying as HTML content.
from xml.dom import minidom

# Use the StringIO function to contain a buffer for the encoded content.

try:

    from cStringIO import StringIO

except ImportError:

    from StringIO import StringIO

import string, types

# Define a special object to indicate when no value is returned from
# executable code.

class NoValueClass:

    pass

NoValue = NoValueClass()


class TypeDict:

    def __init__(self):
    
        self.dict = {}
    
    def resolve_type(self, key):
    
        # Check the type of the key.
        datatype = type(key)
        
        if datatype == types.InstanceType:
        
            # The key is an instance so return the class it was created from.
            return key.__class__
        
        #elif datatype == types.ClassType:
        #
        #    # Return the first base class.
        #    for base in key.__bases__:
        #    
        #        if self.dict.has_key(base):
        #        
        #            return base
        #    
        #    return None
        
        else:
        
            # Return the type found.
            return datatype
    
    def __getitem__(self, key):
    
        # Return the dictionary entry corresponding to the class or type
        # of the key.
        return self.dict[self.resolve_type(key)]
    
    def __setitem__(self, key, value):
    
        # Store the value corresponding to the type given.
        self.dict[key] = value
    
    def has_key(self, key):
    
        return self.dict.has_key(self.resolve_type(key))
    
    def keys(self):
    
        return self.dict.keys()
    
    def values(self):
    
        return self.dict.values()
    
    def items(self):
    
        return self.dict.items()
    
    def clear(self):
    
        return self.dict.clear()
    
    def popitem(self):
    
        return self.dict.popitem()
    
    def __delitem__(self, key):
    
        # Don't resolve the type of the key.
        del self.dict[key]
    
    def get(self, key, default = None):
    
        return self.dict.get(self.resolve_type(key), default)
    
    def setdefault(self, key, default = None):
    
        datatype = self.resolve_type(key)
        
        if self.dict.has_key(datatype):
        
            return self.dict[datatype]
        
        else:
        
            self.dict[datatype] = default
            return default
    
    def __repr__(self):
    
        return repr(self.dict)


# Define a dictionary of formatters. These will be copied to a dictionary
# in the user's workspace module.
__formatters__ = TypeDict()


# Define a function to encapsulate the conversion of strings into an HTML
# compliant encoding.

whitespace = "\t\n\r "

def encode(text, preformatted = 1):

    # Encode the data for use as HTML content.
    t = minidom.Text()
    t.nodeValue = text
    s = StringIO()
    t.writexml(s)
    s.seek(0)
    htmltext = s.read()
    s.close()
    
    # Convert spaces to hard spaces to prevent formatting issues. These
    # occur even within preformatted elements, unfortunately.
    new = []
    
    if preformatted == 1:
    
        for c in htmltext:
        
            if c == " ":
                new.append("&nbsp;")
            elif c == "\n":
                new.append("<br>")
            elif ord(c) < 32 and (c not in whitespace):
                new.append("\\x%02x" % ord(c))
            else:
                new.append(c)
    
    else:
    
        # Use a flag to distinguish between leading whitespace on each line
        # and other occurrences of whitespace in the string.
        leading = 1
        
        for c in htmltext:
        
            if c == "\n":
            
                new.append("<br>")
                
                # Reset the leading flag.
                leading = 1
            
            elif ord(c) < 32 and (c not in whitespace):
            
                # Unprintable ASCII character
                new.append("\\x%02x" % ord(c))
                
                # Clear the leading flag.
                leading = 0
            
            elif ord(c) == 32 and leading == 1:
            
                # Leading space
                new.append("&nbsp;")
            
            else:
            
                new.append(c)
                
                # Clear the leading flag.
                leading = 0
    
    return "".join(new)


# A sample list formatter (as written in a khpython session).

def form_l(l):
 
 n = len(l)
 output = "<table>\n"
 for i in range(0,n):
  output = output + "<tr><td>%s</td>\n" % encode(l[i])
  output = output + "</tr>\n"
 return output + "</table>\n"

# A more readable version (based on the above).
def format_list(l, **kwargs):

    if l == []:
    
        return "<em>[Empty list]</em>"
    
    n = len(l)
    
    output = ['<table border="1" cellpadding="4" class="List">', '<tbody>']
    
    if kwargs.get("orient", "v") == "v":
    
        kwargs["orient"] = "h"
        
        for i in range(0, n):
        
            # Try to format the contents of this element.
            try:
            
                value = __formatters__[l[i]](l[i], **kwargs)
            
            except KeyError:
            
                value = encode(str(l[i]))
            
            output.append("<tr>\n<td>%s</td>\n</tr>\n" % value)
    
    else:
    
        kwargs["orient"] = "v"
        
        output.append("<tr>")
        
        for i in range(0, n):
        
            # Try to format the contents of this element.
            try:
            
                value = __formatters__[l[i]](l[i], **kwargs)
            
            except KeyError:
            
                value = encode(str(l[i]))
            
            output.append("<td>%s</td>\n" % value)
        
        output.append("</tr>")
    
    output = output + ["</tbody>", "</table>"]
    
    return "\n".join(output)

# Add this handler function to the dictionary.
__formatters__[types.ListType] = format_list


# Define a tuple formatter.

def format_tuple(t, **kwargs):

    if t == ():
    
        return "<em>[Empty tuple]</em>"
    
    n = len(t)
    
    output = \
    [
        '<table border="1" cellpadding="4" class="Tuple">', '<tbody>', '<tr>'
    ]
    
    for i in range(0, n):
    
        # Try to format the contents of this element.
        try:
        
            value = __formatters__[t[i]](t[i])
        
        except KeyError:
        
            value = encode(str(t[i]))
        
        output.append("<td>%s</td>\n" % value)
    
    output = output + ["</tr>", "</tbody>", "</table>"]
    
    return "\n".join(output)

# Add this handler function to the dictionary.
__formatters__[types.TupleType] = format_tuple


# Define a formatter for dictionaries.

def format_dict(d, **kwargs):

    if d == {}:
    
        return "<em>[Empty dictionary]</em>"
    
    output = \
    [
        '<table border="1" cellpadding="4" class="Dictionary">', '<tbody>'
    ]
    
    for key, value in d.items():
    
        output.append("<tr>")
        
        # Try to format the contents of this element.
        try:
        
            kvalue = __formatters__[key](key, **kwargs)
        
        except KeyError:
        
            kvalue = encode(str(key))
        
        output.append("<th>%s</th>\n" % kvalue)
        
        try:
        
            vvalue = __formatters__[value](value, **kwargs)
        
        except KeyError:
        
            vvalue = encode(str(value))
        
        output.append("<td>%s</td>\n" % vvalue)
        
        output.append("</tr>")
    
    output = output + ["</tbody>", "</table>"]
    
    return "\n".join(output)

__formatters__[types.DictType] = format_dict


# Define a formatter for the None value which returns no output.

def format_None(value, **kwargs):

    return NoValue

__formatters__[types.NoneType] = format_None


# Define a formatter for the TypeDict class.

def format_TypeDict(td, **kwargs):

    return __formatters__[td.dict](td.dict)

__formatters__[TypeDict] = format_TypeDict


# Define a formatter for file objects.

def format_file(f, **kwargs):

    return '<tt class="File">&lt;File: <a href="file:%s">%s</a>' % \
        (f.name, f.name) + ' (mode "%s")&gt;</tt>' % f.mode

__formatters__[types.FileType] = format_file


# Define a formatter for modules.

def format_module(m, **kwargs):

    return '<tt class="Module">&lt;Module: ' + \
        '<a href="pydoc:%s">%s</a>&gt;</tt>' % (m.__name__, m.__name__)

__formatters__[types.ModuleType] = format_module


# Define a formatter for classes.

def format_class(c, **kwargs):

    return '<tt class="Class">&lt;Class: ' + \
        '<a href="khpython:%s">%s</a>&gt;</tt>' % (c.__name__, c.__name__)

__formatters__[types.ClassType] = format_class
