#!/usr/bin/env python

"""
khpython - an interactive Python shell using PyKDE and a KHTMLPart

(C) David Boddie 2003

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from qt import QFileDialog, QInputDialog, QLineEdit, QMessageBox, \
               QMouseEvent, QObject, \
               QPopupMenu, Qt, QVBox, SIGNAL, SLOT, qApp

from kdecore import \
    i18n, KApplication, KAccel, KAboutData, KCmdLineArgs, KConfig, \
    KKey, KShortcut, KURL

from kdeui import KLineEdit, KMainWindow
from khtml import KHTMLPart
from kparts import KParts

import imp, os, string, sys, traceback

#debug = open("dump.txt", "w")

# Import the __future__ module to enable code using future language changes
# to be compiled interactively.

import __future__


# Import some useful output formatters.
import formatters

# Import the encode function and the appropriate StringIO function which was
# selected in the formatters module.
from formatters import encode, StringIO, NoValue

# Define the application details.

appname = "khpython"
__version__ = "0.71"


# A start up message is defined.

start_up = \
"""
<h1 class = "Title">%s %s</h1>
<p>
  <em>python:</em>
  Type "help", "copyright", "credits" or "license" for more information.
</p>
<p>
  <em>khpython:</em>
  Type "khlicense" for license information specific to this environment.
</p>
""" % (encode(appname), encode(__version__))

# Define a License class.

class License:

    def __init__(self, text):
    
        self.license_text = text

# Define an object to contain the license for khpython.
khlicense = License(__doc__[1:])

# Define a docstring for the user module.
user_docstring = __doc__[1:string.find(__doc__, "\n", 1)]

# Define a simple help function for interactive use (not required).

help_fn = """
def help(word):

    try:
    
        return word.__doc__
    
    except AttributeError:
    
        return "No documentation found."
"""


class Environment:

    def __init__(self):
    
        # Import some modules to customise the environment.
        self.setup_environment()
        
        # Delayed code for execution.
        self.delayed = []
    
    # Provide a cleaner interface to the formatter dictionary for non-user
    # code.
    
    def add_formatter(self, datatype, handler):
    
        self.module.__formatters__[datatype] = handler
    
    def remove_formatter(self, datatype = None, handler = None):
    
        if datatype is not None:
        
            del self.module.__formatters__[datatype]
            return
        
        if handler is not None:
        
            for key, value in self.module.__formatters__.items():
            
                if value == handler:
                
                    del self.module.__formatters__[key]
    
    def setup_environment(self):
    
        # Construct a module to use to contain Python executed by the
        # user.
        self.module = imp.new_module("__user__")
        
        # Add a docstring for this module.
        self.module.__doc__ = user_docstring
        
        # Copy the sys module into the user's environment.
        self.module.sys = sys
        
        # Make a copy in the user module of the dictionary of output
        # formatters in the formatters module.
        self.module.__formatters__ = formatters.__formatters__
        
        # Define some functions for the user.
        
        # Reimplement a help function like that used in Python 2.2.
        
        #help = compile(help_fn, "<setup_functions>", "exec")
        #exec help in self.module.__dict__
        
        # khlicense
        
        # Create a formatter for the License class.
        def format_License(l, **kwargs):
        
            return '<pre class="License">%s</pre>\n' % encode(l.license_text)
        
        self.module.__formatters__[License] = format_License
        
        # Copy the license object into the user module.
        self.module.khlicense = khlicense
        #self.module.LICENSE = __doc__[1:]
        
        # Give the user access to the encode function.
        self.module.encode = encode
    
    def execute(self, command_line):
    
        if len(command_line) == 0:
        
            return NoValue
        
        if command_line[0] in " \t":
        
            # Add the line to the list of delayed command lines.
            self.delayed.append(command_line)
            return NoValue
        
        if self.delayed != []:
        
            command_lines = "".join(self.delayed) + command_line
        
        else:
        
            command_lines = command_line
        
        try:
        
            value = eval(command_lines, self.module.__dict__)
            
            # Set the _ variable to the value returned.
            self.module._ = value
            
            # The pending input was executed successfully so clear
            # the delayed command lines list.
            self.delayed = []
            
            # Use the appropriate formatter to present the value
            # obtained.
            try:
            
                return self.module.__formatters__[value](value)
            
            except KeyError:
            
                return '<tt class = "Value">%s</tt>' % encode(str(value), preformatted = 0)
        
        except SyntaxError:
        
            # Catch Syntax Errors and pass them on for another attempt
            # to compile and execute them.
            pass
        
        except:
        
            # Catch all other errors and report them.
            
            # Clear the delayed command line list.
            self.delayed = []
            return self.format_exc()
        
        
        # Try again with the exec command.
        
        # Construct a flags word for any features imported from the
        # __future__ module. Peculiarly, this has to be done because
        # the compile function appears not to be aware of any __future__
        # imports done from the command line.
        
        flags = 0
        user_values = filter(
            lambda x: isinstance(x, __future__._Feature),
            self.module.__dict__.values()
            )
        
        for value in __future__.__dict__.values():
        
            if isinstance(value, __future__._Feature):
            
                if value in user_values:
                
                    flags = flags | value.compiler_flag
        
        try:
        
            # Compile the code using the flags determined from the imported
            # features. The last parameter indicates that these flags should
            # be taken into account when compiling.
            
            code = compile(command_lines, "__user__", "exec", flags, 0)
            exec code in self.module.__dict__
            
            # The pending input was executed successfully so clear
            # the delayed command lines list.
            self.delayed = []
            
            # No values are returned from the execution of code in this
            # manner unless we are using existing lines from the delayed
            # lines list.
            return NoValue
        
        except SyntaxError:
        
            msg = sys.exc_info()[1].msg
            #print encode(msg)
            
            if string.find( msg, "unexpected EOF while parsing" ) != -1:
            
                # Delay code for later execution.
                self.delayed.append(command_line)
                return NoValue
            
            elif string.find(msg, "expected an indented block") != -1:

                # Only take this seriously if there is no preceding text
                # or if these last two lines have the same indentation.
                
                if self.delayed != [] and \
                    ( self.indentation(self.delayed[-1]) != \
                      self.indentation(command_line)          ):
                
                    # Keep the line for later if there is preceding text and
                    # the indentation has changed.
                    
                    self.delayed.append(command_line)
                    return NoValue
            
            # An actual Syntax Error.
            
            # Clear the delayed command line list.
            self.delayed = []
            return self.format_exc()
        
        #except IndentationError:
        #
        #    # Only take this seriously if there is no preceding text.
        #    
        #    if self.delayed != []:
        #    
        #        self.delayed.append(command_line)
        #    
        #    else:
        #    
        #        return self.format_exc()
        
        except:
        
            # Other errors are caught here.
            
            # Clear the delayed command line list.
            self.delayed = []
            return self.format_exc()
    
    def format_exc(self):
    
        lines = map(
            lambda x: encode(x),
            #traceback.format_exception_only(*(sys.exc_info()[:2]))
            traceback.format_exception(*(sys.exc_info()))
            )
        
        return '<p class = "Error">' + "<br>\n".join(lines) + '</p>\n'
    
    def indentation(self, s, tabsize = 4):
    
        l = 0
        
        for i in s:
        
            if i == " ":
                l = l + 1
            elif i == "\t":
                l = l + tabsize
            else:
                break
        
        return l


class Buffer:

    def __init__(self, *args):
    
        self.StringIO = StringIO(*args)
        self.read_callback = self.null_read_callback
        self.write_callback = self.null_write_callback
    
    def null_read_callback(self, buf, data):
    
        return data
    
    def null_write_callback(self, buf, data, flush = 0):
    
        pass
    
    def write(self, data):
    
        # Use another method to actually write the data without triggering
        # a callback.
        self._write(data)
        
        # Call callback.
        self.write_callback(self, data)
    
    def _write(self, data):
    
        self.StringIO.write(data)
    
    def read(self, *args):
    
        # Use another method to actually read the data without triggering
        # a callback.
        data = self._read(*args)
        
        # Call callback with the data in the buffer, giving the callback
        # the chance to modify it before its use.
        data = self.read_callback(self, data)
        
        return data
    
    def _read(self, *args):
    
        return self.StringIO.read(*args)
    
    def clear(self):
    
        self.StringIO.seek(0, 0)
        self.StringIO.truncate()
    
    def flush(self):
    
        self.write_callback(self, "", flush = 1)
        self.StringIO.flush()
#    
#    def isatty(self):
#    
#        self.StringIO.isatty()
#    
#    def getvalue(self, *args):
#    
#        self.StringIO.getvalue(*args)
#    
    def readline(self):
    
        # Use another method to actually read the data without triggering
        # a callback.
        data = self._readline()
        
        # Call callback with the data in the buffer, giving the callback
        # the chance to modify it before its use.
        data = self.read_callback(self, data)
        
        return data
    
    def _readline(self):
    
        return self.StringIO.readline()
    
    def readlines(self):
    
        # Use another method to actually read the data without triggering
        # a callback.
        lines = self._readlines()
        
        new_lines = []
        for line in lines:
        
            # Call callback.
            new_line = self.read_callback(self, data[:-1]) + "\n"
            new_lines.append(new_line)
        
        return new_lines
    
    def _readlines(self):
    
        return self.StringIO.readlines()
    
    def seek(self, *args):
    
        self.StringIO.seek(*args)
    
    def tell(self):
    
        return self.StringIO.tell()
    
    def writelines(self, *args):
    
        # Use another method to actually write the data without triggering
        # a callback.
        self._writelines(*args)
        
        # Call callback.
        self.write_callback(self, "\n".join(args))
    
    def _writelines(self, *args):
    
        self.StringIO.writelines(*args)
    
    def truncate(self, *args):
    
        if args != ():
        
            self.StringIO.seek(args[0])
        
        self.StringIO.truncate()


class HTMLPart(KHTMLPart):

    def __init__(self, layout, parent):
    
        KHTMLPart.__init__(self, layout)
        #QObject.connect(self, SIGNAL("urlSelected(
        
        # Record a reference to the parent object.
        self.parent = parent
        
        # Set up a list to record the source text.
        self.source_list = []
    
    def urlSelected(self, url, button, state, target, args):
    
        if button == QMouseEvent.LeftButton:
        
            # Construct a new URL.
            new_url = self.make_url(url)
            
            # Open the URL.
            self.parent.openURL(new_url)
    
    def make_url(self, url):
    
        # Is the URL absolute or relative?
        if KURL.isRelativeURL(url):
        
            # A relative URL
            
            # Join the current URL and the new URL.
            new_url = KURL(KURL(self.url()), url)
        
        else:
        
            # Create a KURL object to use to open the new URL.
            new_url = KURL(url)
        
        return new_url
        
    def decode_string(self, name):
    
        new = u""
        
        i = 0
        
        while i < len(name):
        
            c = name[i:i+3]
            
            if c == u"%2f":
            
                new = new + u"/"
                i = i + 3
            
            elif c == u"%25":
            
                new = new + u"%"
                i = i + 3
            
            elif c == u"%23":
            
                new = new + u"#"
                i = i + 3
            
            else:
            
                new = new + c[:1]
                i = i + 1
        
        # There should be no trailing % symbols.
        return new
    
    def begin(self):
    
        KHTMLPart.begin(self)
        
        # Reset the source list.
        self.source_list = []
    
    def write(self, text):
    
        KHTMLPart.write(self, text)
        
        # Record the text for possible future use.
        self.source_list.append(text)
    
    def source(self):
    
        return string.join(self.source_list, "")
    
    def page_up(self):
    
        # Determine the height of the view onto the HTML.
        height = self.view().height()
        
        # Scroll the view upwards by this amount.
        self.view().scrollBy(0, height)
    
    def page_down(self):
    
        # Determine the height of the view onto the HTML.
        height = self.view().height()
        
        # Scroll the view downwards by this amount.
        self.view().scrollBy(0, -height)


class LineEdit(KLineEdit):

    def __init__(self, parent, slot):
    
        KLineEdit.__init__(self, parent)
        
        QObject.connect(
            self, SIGNAL("returnPressed()"), self.enter_line
            )
        
        #self.useGlobalKeyBindings()
        
        self.slot = slot
        
        self.history = []
        self.history_pos = len(self.history)
    
    def history_previous(self):
    
        self.history_pos = max(0, self.history_pos - 1)
        
        if self.history_pos < len(self.history):
        
            previous = self.history[self.history_pos]
        
        else:
        
            previous = ""
        
        self.setText(previous)
    
    def history_next(self):
    
        items = len(self.history)
        
        self.history_pos = min(self.history_pos + 1, items)
        
        if self.history_pos < items:
        
            next = self.history[self.history_pos]
        
        else:
        
            next = ""
        
        self.setText(next)
    
    def enter_line(self):
    
        self.history.append(self.text())
        self.history_pos = len(self.history)
        
        # Call the slot passed by the parent object.
        self.slot()


class HTMLWindow(KMainWindow):

    # Redefine the tr() function for this class.
    def tr(self, text):
    
        return qApp.translate("HTMLWindow", text)
    
    def __init__(self, stylesheet = ""):
    
        KMainWindow.__init__(self)
        
        # Create a layout widget to use to contain the HTML display and
        # the prompt widget.
        self.vbox = QVBox(self)
        self.setCentralWidget(self.vbox)
        
        self.htmlpart = HTMLPart(self.vbox, self)
        #self.htmlpart.view().reparent(self, QPoint(0, 0), 1)
        
        self.prompt = LineEdit(self.vbox, self.enter_line)
        
        # Read the configuration file.
        self.load_config()
        
        # Create an environment to run Python scripts in.
        self.environment = Environment()
        
        # Redirect stdin, stdout and stderr to suitable objects.
        self.old_stdin = sys.__stdin__
        stdin = Buffer()
        stdin.read_callback = self.process_stdin
        sys.__stdin__ = sys.stdin = stdin
        
        self.old_stdout = sys.__stdout__
        stdout = Buffer()
        stdout.write_callback = self.process_stdout
        sys.__stdout__ = sys.stdout = stdout
        
        self.old_stderr = sys.__stderr__
        stderr = Buffer()
        stderr.write_callback = self.process_stderr
        sys.__stderr__ = sys.stderr = stderr
        
        # Add an output stream for writing directly to the page.
        self.environment.module.htmlout = Buffer()
        self.environment.module.htmlout.write_callback = self.process_htmlout
        
        # Add a new page function to the environment module.
        self.environment.module.new_page = self.new_page
        
        # Add a save output function to the environment module.
        self.environment.module.save_page = self.save_page
        
        # Start writing output to the HTML document.
        self.new_page(stylesheet)
        
        # Print a welcome message.
        self.add_output(start_up)
        
        # Create menus for the application (making the window the parent
        # will prevent the Qt object from being deleted when the Python
        # object goes out of scope).
        file_menu = QPopupMenu(self)
        file_menu.insertItem(
            self.tr("&Save page"), self.open_save_dialog, self.tr("Ctrl+S")
            )
        file_menu.insertItem(
            self.tr("E&xit"), qApp, SLOT("quit()"), self.tr("Ctrl+Q")
            )
        
        # Insert the file menu into the window's menu bar.
        self.menuBar().insertItem(self.tr("&File"), file_menu)
        
        # Use a counter to generate anchors so that the view can scroll.
    
    def __del__(self):
    
        sys.__stdin__ = sys.stdin = self.old_stdin
        sys.__stdout__ = sys.stdout = self.old_stdout
        sys.__stderr__ = sys.stderr = self.old_stderr
        #debug.write("HTMLWindow.__del__%s\n" % repr((self)))
        
        #self.htmlpart.view().reparent(None, QPoint(0, 0), 1)
        KMainWindow.__del__(self)
    
    def load_config(self):
    
        config = KConfig("khpython")
        
        config.setGroup("Geometry")
        
        x, y = None, None
        
        if config.hasKey("Default horizontal position") and \
            config.hasKey("Default vertical position"):
            
            x = config.readNumEntry("Default horizontal position")
            y = config.readNumEntry("Default vertical position")
        
        if x is not None:
        
            self.move(x, y)
        
        w, h = 512, 512
        
        if config.hasKey("Default window width") and \
            config.hasKey("Default window height"):
        
            w = config.readNumEntry("Default window width")
            h = config.readNumEntry("Default window height")
        
        self.resize(w, h)
        
        # Read keyboard shortcuts.
        config.setGroup("Shortcuts")
        
        # Create an accelerator for the window as a whole.
        
        self.main_keys = KAccel(self)
        
        self.main_keys.insert(
            "Exit", i18n("Exit"),
            i18n("This action causes the application to terminate."),
            KShortcut(config.readEntry("Exit", "CTRL+q")),
            self.close
            )
        
        # Create an accelerator for the line edit widget used as a prompt.
        
        self.line_keys = KAccel(self.prompt)
        
        self.line_keys.insert(
            "Previous line", i18n("Previous line"),
            i18n("Retrieve the previous line in the command history buffer."),
            KShortcut(config.readEntry("Previous line in history", "Up")),
            self.prompt.history_previous
            )
        
        self.line_keys.insert(
            "Next line", i18n("Next line"),
            i18n("Retrieve the next line in the command history buffer."),
            KShortcut(config.readEntry("Next line in history", "Down")),
            self.prompt.history_next
            )
        
        self.line_keys.insert(
            "Page up", i18n("Page up"),
            i18n("Scroll the HTML view up by a page height."),
            KShortcut(config.readEntry("Page up", "PageUp")),
            self.htmlpart.page_up
            )
        
        self.line_keys.insert(
            "Page down", i18n("Page down"),
            i18n("Scroll the HTML view down by a page height."),
            KShortcut(config.readEntry("Page down", "PageDown")),
            self.htmlpart.page_down
            )
        
        
        # Read appearance settings.
        config.setGroup("Appearance")
        
        # Default style sheet to use.
        self.stylesheet = str(config.readEntry("Stylesheet", ""))
        
        if self.stylesheet == "":
        
            # Find the absolute path to this program, and hence the default
            # stylesheet.
            app_abspath = os.path.abspath(app_path)
            
            self.stylesheet = "file:" + os.path.join(app_abspath, "styles.css")
        
        if config.hasKey("Display width"):
        
            self.display_width = int(config.readNumEntry("Display width"))
        
        else:
        
            self.display_width = 80
    
    def close(self, alsoDelete = 1):
    
        #debug.write("close%s\n" % repr((self, alsoDelete)))
        
        self.htmlpart.end()
        KMainWindow.close(self, 1)
    
    def queryClose(self):
    
        #debug.write("queryClose\n")
        return 1
    
    def queryExit(self):
    
        #debug.write("queryExit\n")
        self.save_config()
        return 1
    
    def quit(self):
    
        self.save_config()
    
    def save_config(self):
    
        config = KConfig("khpython")
        
        config.setGroup("Geometry")
        
        x, y = self.x(), self.y()
        
        config.writeEntry("Default horizontal position", x)
        config.writeEntry("Default vertical position", y)
        
        w, h = self.width(), self.height()
        
        config.writeEntry("Default window width", w)
        config.writeEntry("Default window height", h)
        
        self.main_keys.setConfigGroup("Shortcuts")
        self.line_keys.setConfigGroup("Shortcuts")
        
        self.main_keys.writeSettings(config)
        self.line_keys.writeSettings(config)
        
        config.setGroup("Appearance")
        
        config.writeEntry("Stylesheet", self.stylesheet)
        config.writeEntry("Display width", self.display_width)
        
        config.sync()
    
    def openURL(self, url):
    
        if unicode(url).find(u"khpython:") == 0:
        
            # Open a window containing details of the object specified.
            pass
        
        else:
        
            # Open the URL in the configured browser.
            self.run = KParts.BrowserRun(
                KURL(url), KParts.URLArgs(), None, None, 1, 0
                )
    
    def new_page(self, stylesheet = ""):
    
        self.htmlpart.begin()
        
        self.add_output(
            "<html>\n<head>\n<title>%s %s</title>\n" % (
                encode(appname), encode(__version__)
                )
            )
        
        self.add_output(
            '<meta http-equiv="Content-Type" content="text/html">\n'
            )
        
        # Set the stylesheet.
        if stylesheet != "":
        
            self.stylesheet = stylesheet
        
        if self.stylesheet != "":
        
            self.add_output(
                '<link rel="stylesheet" type="text/css" ' + \
                'href="%s">\n' % encode(self.stylesheet)
                )
            #try:
            #
            #    stylesheet_text = open(stylesheet, "rb").read()
            #    #self.htmlpart.setUserStyleSheet(KURL(stylesheet))
            #    
            #    self.htmlpart.preloadStyleSheet(
            #        stylesheet, QString(stylesheet_text)
            #        )
            #
            #except IOError:
            #
            #    pass
        
        # End the header.
        self.add_output("</head>\n<body>\n")
    
    def enter_line(self):
    
        text = str(self.prompt.text())
        
        # Echo the text submitted to the output window.
        #self.htmlpart.begin()
        self.add_output('<tt class = "Input">%s</tt><br>\n' % encode(text))
        # Examine the text submitted.
        
        # Submit it to the user's running environment.
        try:
        
            # Retrieve a string value from the environment's execute method.
            value = self.environment.execute(text+"\n")
            
            # If no exception was raised then clear the prompt text.
            self.prompt.clear()
            
            if value != NoValue:
            
                self.add_output(value+"<br>\n")
        
        except SyntaxError:
        
            pass
        
        #self.htmlpart.end()
    
    def add_output(self, output):
    
        self.htmlpart.write(output)
        self.htmlpart.view().layout()
        
        # Examine the height of the scroll area and ensure that the output
        # is visible.
        width = self.htmlpart.view().contentsWidth()
        height = self.htmlpart.view().contentsHeight()
        x = self.htmlpart.view().contentsX()
        y = self.htmlpart.view().contentsY()
        vwidth = self.htmlpart.view().visibleWidth()
        vheight = self.htmlpart.view().visibleHeight()
        self.htmlpart.view().ensureVisible(0, height)
        #self.htmlpart.view().scrollBy(0, height - y - vheight)
        #self.htmlpart.view().updateContents()
    
    def open_save_dialog(self, *args):
    
        """open_save_dialog(self, *args)
        
        Open a save dialog to allow the user to specify a file to contain
        the HTML page output.
        """
        
        filename = QFileDialog.getSaveFileName(
            "", self.tr("HTML (*.html)"), self, "save HTML",
            self.tr("Save the page as:")
            )
        
        if not filename:
        
            QMessageBox.warning(
                self, self.tr("No filename given"),
                self.tr("Please specify the file to write."),
                QMessageBox.Cancel, QMessageBox.NoButton, QMessageBox.NoButton
                )
        
        else:
        
            self.save_page(unicode(filename))
        
    def save_page(self, filename):
    
        """save_page(filename)
        
        Save the HTML output to the file specified.
        """
        
        # Write a terminated version of the HTML.
        
        try:
        
            open(filename, "w").write(
                self.htmlpart.source() + "</body>\n</html>\n"
                )
        
        except IOError:
        
            QMessageBox.warning(
                self, self.tr("Cannot save page"),
                self.tr("The page could not be saved to the specified file."),
                QMessageBox.Cancel, QMessageBox.NoButton,
                QMessageBox.NoButton
                )
    
    def end_output(self):
    
        self.htmlpart.end()
    
    def process_stdin(self, buf, data):
    
        # Flush stdout and stderr before obtaining input.
        sys.stdout.flush()
        sys.stderr.flush()
        
        new_data, ok = QInputDialog.getText(
            appname, "Request for input", QLineEdit.Normal, data,
            self
            )
        
        if ok == 1:
        
            # Echo stdin on stdout.
            sys.stdout.write(str(new_data)+"\n")
            
            return str(new_data)+"\n"
        
        else:
        
            return str(data)
    
    def process_output(self, buf, data, flush):
    
        # Take the data recently added to the buffer and add it to the
        # HTML content.
        
        if flush == 1 or data[-1:] == "\n":
        
            buf.seek(0, 0)
            
            # Use the public read method to extract the actual data.
            data = buf.read()
            
            #if flush == 0 and (data[:-1] == "\n" and data != "\n"):
            #
            #    # Remove trailing newlines from strings which contain
            #    # other characters, leaving single newlines alone.
            #    data = data[:-1]
            
            # Reformat, taking carriage return characters into account.
            lines = string.split(data, "\n")
            
            new_lines = []
            
            for line in lines:
            
                # Split the line at each carriage return encountered.
                words = string.split(line, "\r")
                
                # Construct a new line by overlaying all the "words"
                # which make up the line.
                length = reduce(lambda x,y: max(x, len(y)), words, 0)
                new_line = " " * min(self.display_width, length)
                
                for word in words:
                
                    # Simply overlay the word over the current line string.
                    # This is not optimal if the words are very long; in that
                    # case they should be wrapped at a specified column number.
                    new_line = word + new_line[len(word):]
                
                new_lines.append(new_line)
            
            data = string.join(new_lines, "\n")
            
            return encode(str(data), preformatted = 0)
        
        return None
    
    def process_stdout(self, buf, data, flush = 0):
    
        output = self.process_output(buf, data, flush)
        
        if output is not None:
        
            # Use the appropriate formatter to present the value
            # obtained.
            
            self.add_output('<tt class = "stdout">%s</tt>' % output)
            
            # Clear the buffer.
            buf.clear()
    
    def process_stderr(self, buf, data, flush = 0):
    
        output = self.process_output(buf, data, flush)
        
        if output is not None:
        
            # Use the appropriate formatter to present the value
            # obtained.
            
            self.add_output('<tt class = "stderr">%s</tt>' % output)
            
            # Clear the buffer.
            buf.clear()
    
    def process_htmlout(self, buf, data, flush = 1):
    
        buf.seek(0, 0)
        
        # Use the public read method to extract the actual data.
        data = buf.read()
        
        self.add_output(data)
        
        # Clear the buffer.
        buf.clear()
    

def main():

    if len(sys.argv) == 2:
    
        stylesheet = sys.argv[1]
    
    else:
    
        stylesheet = ""
    
    # Find the path to this program.
    global app_path, app_file
    
    app_path, app_file = os.path.split(sys.argv[0])
    
    # If the working directory is not in the sys.path list then add it.
    
    pwd = os.getenv("PWD")
    
    if pwd != "" and pwd not in sys.path:
    
        sys.path.insert(0, pwd)
    
    aboutData = KAboutData(app_file, "KHTML Python shell", __version__)
    
    KCmdLineArgs.init(sys.argv[:1], aboutData)
    
    # Create a new application.
    if sys.argv[1:] == []:
    
        args = [""]
    
    else:
    
        args = sys.argv[1:]
    
    #app = KApplication(args, appname)
    app = KApplication()
    
    
    # Create a main window.
    main = HTMLWindow(stylesheet)
    
    # Show the window.
    main.show()
    
    # Set the application's main widget.
    app.setMainWidget(main)
    
    # Handle events.
    app.exec_loop()


if __name__ == "__main__":

    if len(sys.argv) > 2:
    
        print "Usage: %s [stylesheet]" % sys.argv[0]
        sys.exit()
    
    main()
    
    # Exit
    sys.exit()
